from itertools import tee
from typing import List
from typing import Text


def ngrams(sequence: List[Text], n: int, min_length: int = 5):
    """
    Return the ngrams generated from a sequence of items, as an iterator.

    This is a modified version of nltk.util.ngrams.

    Parameters
    ----------
    sequence : List[Text]
        The sequence of items.
    n : int
        The length of each ngram.
    min_length : int, optional
        The minimum length of each ngram, by default 5

    Returns
    -------
    iterator
        The ngrams.

    Examples
    --------
    >>> list(ngrams(["a", "b", "c", "d"], 2, min_length=1))
    [('a', 'b'), ('b', 'c'), ('c', 'd')]
    >>> list(ngrams(["a", "b", "c", "d"], 2, min_length=5))
    []
    >>> list(ngrams(["a", "b"], 3, min_length=1))
    [('a', 'b')]
    """
    if len(sequence) < min_length:
        return []
    if len(sequence) < n:
        return [tuple(sequence)]
    iterables = tee(iter(sequence), n)
    for i, sub_iterable in enumerate(iterables):
        for _ in range(i):
            next(sub_iterable, None)
    return zip(*iterables)


import regex as re

DIGIT_RE = re.compile(r"\d")
PUNCT_OR_NON_PRINTING_CHARS_RE = re.compile(r"[\p{P}\p{C}\p{S}]+")


def normalize(line: str) -> str:
    """
    Normalize a line of text. Source: https://github.com/facebookresearch/cc_net/blob/bda555bd1cf1ee2e0b925363e62a61cd46c8b60d/cc_net/text_normalizer.py#L180

    Parameters
    ----------
    line : str
        The line of text to normalize.

    Returns
    -------
    str
        The normalized line of text.

    Examples
    --------
    >>> normalize("Hello, world!")
    'hello world'
    >>> normalize("Hello, 123!\\n\\t\\b")
    'hello 000'
    """
    line = line.strip()
    if not line:
        return line
    line = line.lower()
    line = DIGIT_RE.sub("0", line)
    line = PUNCT_OR_NON_PRINTING_CHARS_RE.sub("", line)
    return line


from typing import List

from scipy.integrate import quad as integrate

from clean_text_my.text_dedup.utils.tokenization import ngrams


def jaccard_similarity(
    doc1: str | List[str],
    doc2: str | List[str],
    ngram_size: int = 8,
    min_length: int = 0,
) -> float:
    """Compute the Jaccard similarity between two documents.

    Parameters
    ----------
    doc1 : str or List[str]
        The first document.
    doc2 : str or List[str]
        The second document.
    ngram_size : int, optional
        The size of n-grams, by default 8
    min_length : int, optional
        The minimum length of each n-gram, by default 0

    Returns
    -------
    float
        The Jaccard similarity.

    Examples
    --------
    >>> jaccard_similarity("hello world", "hello world")
    1.0
    >>> jaccard_similarity("hello world", "hello world!")
    0.8
    >>> jaccard_similarity("hello world".split(), "hello world!".split(), ngram_size=1)
    0.3333333333333333
    """
    words1 = set(" ".join(ng) for ng in ngrams(list(doc1), ngram_size, min_length=min_length))
    words2 = set(" ".join(ng) for ng in ngrams(list(doc2), ngram_size, min_length=min_length))
    return len(words1 & words2) / max(1, len(words1 | words2))


def optimal_param(
    threshold: float,
    num_perm: int,
    false_positive_weight: float = 0.5,
    false_negative_weight: float = 0.5,
):
    """
    Compute the optimal `MinHashLSH` parameter that minimizes the weighted sum
    of probabilities of false positive and false negative, taken from datasketch.

    You can also refer to the interactive demo at https://huggingface.co/spaces/bigcode/near-deduplication.

    Parameters
    ----------
    threshold : float
        The threshold for similarity.
    num_perm : int
        The number of permutations.
    false_positive_weight : float
        The weight of false positive.
    false_negative_weight : float
        The weight of false negative.

    Returns
    -------
    Tuple[int, int]
        The optimal `b` (bands) and `r` (rows) parameters.

    Examples
    --------
    >>> optimal_param(0.75, 256)
    (21, 12)
    >>> optimal_param(0.75, 256, 0.1, 0.9)
    (28, 9)
    """

    def false_positive_area(threshold: float, b: int, r: int):
        """Source: `datasketch.lsh`"""

        def proba(s):
            return 1 - (1 - s ** float(r)) ** float(b)

        a, _ = integrate(proba, 0.0, threshold)
        return a

    def false_negative_area(threshold: float, b: int, r: int):
        """Source: `datasketch.lsh`"""

        def proba(s):
            return 1 - (1 - (1 - s ** float(r)) ** float(b))

        a, _ = integrate(proba, threshold, 1.0)
        return a

    min_error = float("inf")
    opt = (0, 0)
    for b in range(1, num_perm + 1):
        max_r = int(num_perm / b)
        for r in range(1, max_r + 1):
            fp = false_positive_area(threshold, b, r)
            fn = false_negative_area(threshold, b, r)
            error = fp * false_positive_weight + fn * false_negative_weight
            if error < min_error:
                min_error = error
                opt = (b, r)
    return opt


import os
import json
from glob import glob
from tqdm import tqdm
from clean_text_my.logging import logger
from clean_text_my.utils import write_to_json, is_additional_slash, create_dir


def data_flattening(
    output_folder: str,
    input_folder: str = None,
    input_filename: str = None,
    custom_key: list = None,
):
    suggested_keys = [
        "p",
        "text",
        "article_text",
        "article_body",
        "text",
        "content",
        "contents",
        "body",
        "articleBody",
        "data",
        "title",
    ]

    output_folder = is_additional_slash(output_folder)
    create_dir(output_folder)

    if input_filename and input_folder:
        raise Exception(
            f"please choose only one option: input_filename or input_folder"
        )

    if custom_key:
        suggested_keys = list(set(suggested_keys + custom_key))

    if input_folder:
        input_folder = is_additional_slash(input_folder)
        files_lst = glob(f"{input_folder}*.jsonl")
    elif input_filename:
        print(input_filename)
        if input_filename.endswith("jsonl"):
            files_lst = [input_filename]
        else:
            raise Exception(f"input_filename is not .jsonl format")
    else:
        raise Exception(f"input_folder and input_filename is none")

    txt_lst = []

    for f in files_lst:
        with open(f) as fopen:
            data = [json.loads(line) for line in fopen]

        try:
            key_data = [key for key, _ in data[0].items()]
            logger.info(f"availble key - {key_data}")
        except AttributeError:
            logger.error(f"dataset not in standard format, assumed got nested array.")
            continue

        if not any(key in key_data for key in suggested_keys):
            logger.error(
                f"dataset not in standard key-value. must have ({' | '.join(suggested_keys)})"
            )
            continue

        for i in tqdm(data):
            str_lst = []
            for key in i.keys():
                if key in suggested_keys:
                    str_lst.append(str(i[key]))
                else:
                    continue

            if None in str_lst:
                str_lst = ["None" if v is None else v for v in str_lst]

            str_data = "\n\n".join(str_lst)
            txt_lst.append({"text": f"{str_data}"})

        _, filename = os.path.split(f)
        write_to_json(txt_lst, f"{output_folder}{filename}")

        logger.info(f"Data flattening completed for {filename}")



import os
import re
import json
import subprocess
from tqdm import tqdm
from pathlib import Path
from sys import platform


def platform_name() -> str:
    if platform == "linux" or platform == "linux2" or platform == "darwin":
        return "linux"
    elif platform == "win32":
        return "windows"


def is_dir(path: str) -> bool:
    return os.path.isdir(path)


def is_endwith_slash(path: str) -> bool:
    if platform_name() == "windows":
        return path.endswith("\\")
    elif platform_name() == "linux":
        return path.endswith("/")


def slash() -> str:
    if platform_name() == "windows":
        return "\\"
    elif platform_name() == "linux":
        return "/"


def is_additional_slash(path) -> str:
    if not is_endwith_slash(path):
        path = f"{path}{slash()}"

    return path


def create_dir(path: str) -> None:
    Path(path).mkdir(parents=True, exist_ok=True)


def write_to_json(data_list: list, filename: str) -> None:
    with open(filename, "w+") as file:
        for item in tqdm(data_list):
            x = json.dumps(item, ensure_ascii=False)
            file.write(x + "\n")



class UnionFind:
    """
    A data structure for maintaining disjoint sets. This helps build connected components for given duplicate pairs.

    Examples
    --------
    >>> uf = UnionFind()
    >>> uf.union(1, 2)
    >>> uf.union(2, 3)
    >>> uf.union(4, 5)
    >>> uf.find(1)
    1
    >>> uf.find(2)
    1
    >>> uf.find(3)
    1
    >>> uf.find(4)
    4
    >>> uf.find(5)
    4
    """

    def __init__(self):
        self.parent = {}

    def find(self, x):
        if x not in self.parent:
            self.parent[x] = x
            return x

        if self.parent[x] != x:
            self.parent[x] = self.find(self.parent[x])

        return self.parent[x]

    def union(self, x, y):
        px = self.find(x)
        py = self.find(y)
        self.parent[px] = self.parent[py] = min(px, py)


import hashlib
import struct
from hashlib import md5
from hashlib import sha256

import xxhash
from xxhash import xxh3_64
from xxhash import xxh3_64_digest
from xxhash import xxh3_128
from xxhash import xxh3_128_digest


def md5_hexdigest(data: bytes) -> str:
    """
    Generate a md5 hex hash from the given data.

    Parameters
    ----------
    data : bytes
        The data to be hashed.

    Returns
    -------
    str
        The hex hash value.

    Examples
    --------
    >>> md5_hexdigest(b"hello world")
    '5eb63bbbe01eeed093cb22bb8f5acdc3'
    >>> len(md5_hexdigest(b"hello world"))
    32
    """
    return md5(data).hexdigest()


def sha1_hash(data: bytes, d: int = 32) -> int:
    """
    Generate a d-bit hash value from the given data.

    Parameters
    ----------
    data : bytes
        The data to be hashed.
    d : int
        The number of bits of the hash value.

    Returns
    -------
    int
        The hash value.

    Examples
    --------
    >>> sha1_hash(b"hello world", 32)
    896314922
    >>> sha1_hash(b"hello world", 64)
    13028719972609469994
    >>> sha1_hash(b"hello world", 128)
    310522945683037930239412421226792791594
    """
    if d == 32:
        return struct.unpack("<I", hashlib.sha1(data).digest()[:4])[0]
    if d == 64:
        return struct.unpack("<Q", hashlib.sha1(data).digest()[:8])[0]
    # struct is faster but does not support arbitrary bit lengths
    return int.from_bytes(hashlib.sha1(data).digest()[: d // 8], byteorder="little")


def sha256_hexdigest(data: bytes) -> str:
    """
    Generate a sha256 hex hash from the given data.

    Parameters
    ----------
    data : bytes
        The data to be hashed.

    Returns
    -------
    str
        The hex hash value.

    Examples
    --------
    >>> sha256_hexdigest(b"hello world")
    'b94d27b9934d3e08a52e52d7da7dabfac484efe37a5380ee9088f7ace2efcde9'
    >>> len(sha256_hexdigest(b"hello world"))
    64
    """
    return sha256(data).hexdigest()


def xxh3_16hash(data: bytes, seed: int = 0) -> int:
    """
    Generate a 16-bit xxhash based hash value from the given data.
    As of python xxhash 3.3.0 (and since 0.3.0) outputs in big-endian.
    This is useful as a special purpose xxhash when you only want 16 bits.
    bit masked xxh3_64 hashes are faster than xxh32 in modern systems.

    Parameters
    ----------
    data : bytes
        The data to be hashed.
    seed : int
        xxhashes can all be seeded. Default is int=0

    Returns
    -------
    int
        The hash value.

    Examples
    --------
    >>> xxh3_16hash(b"hello world")
    39051
    >>> xxh3_16hash(b"hello world",seed=42)
    13198
    >>> xxh3_16hash(b"hello world",seed=-42)
    34281
    """
    return xxhash.xxh3_64_intdigest(data, seed) & 0xFFFF


def xxh3_32hash(data: bytes, seed: int = 0) -> int:
    """
    Generate a 32-bit xxhash based hash value from the given data.
    As of python xxhash 3.3.0 (and since 0.3.0) outputs in big-endian.
    This is useful as a special purpose xxhash when you only want 32bits.
    bit masked xxh3_64 hashes are faster than xxh32 in modern systems.

    Parameters
    ----------
    data : bytes
        The data to be hashed.
    seed : int
        xxhashes can all be seeded. Default is int=0

    Returns
    -------
    int
        The hash value.

    Examples
    --------
    >>> xxh3_32hash(b"hello world")
    1088854155
    >>> xxh3_32hash(b"hello world",seed=42)
    3913102222
    >>> xxh3_32hash(b"hello world",seed=-42)
    3721037289
    """
    return xxhash.xxh3_64_intdigest(data, seed) & 0xFFFFFFFF


def xxh3_hash(data: bytes, d: int = 32) -> int:
    """
    Generate a d-bit xxhash based hash value from the given data.
    As of python xxhash 3.3.0 (and since 0.3.0) outputs in big-endian.
    This is useful as a general purpose xxhash that can take multiple `d` values

    Parameters
    ----------
    data : bytes
        The data to be hashed.
    d : int
        The number of bits of the hash value.
        According to this value, chooses empirically found best xxh3 hasher.

    Returns
    -------
    int
        The hash value.

    Examples
    --------
    >>> xxh3_hash(b"hello world", 32)
    1088854155
    >>> xxh3_hash(b"hello world", 64)
    15296390279056496779
    >>> xxh3_hash(b"hello world", 128)
    297150157938599054391163723952090887879
    """
    match d:
        case 32:
            # with sse2 or later, xxh3 is much faster
            # with avx, the difference is much larger
            return xxhash.xxh3_64_intdigest(data) & 0xFFFFFFFF
        case 64:
            return xxhash.xxh3_64_intdigest(data)
        case 128:
            return xxhash.xxh3_128_intdigest(data)
    # fall back
    return int.from_bytes(xxhash.xxh3_128_digest(data)[: d // 8], byteorder="big")


__all__ = [
    "md5",
    "sha256",
    "sha1_hash",
    "xxh3_64",
    "xxh3_64_digest",
    "xxh3_128",
    "xxh3_128_digest",
    "xxh3_hash",
    "xxh3_16hash",
    "xxh3_32hash",
]


# from text_dedup.utils.add_args import add_bloom_filter_args
# from text_dedup.utils.add_args import add_exact_hash_args
# from text_dedup.utils.add_args import add_io_args
# from text_dedup.utils.add_args import add_meta_args
# from text_dedup.utils.add_args import add_minhash_args
# from text_dedup.utils.add_args import add_sa_args
# from text_dedup.utils.add_args import add_simhash_args
# from text_dedup.utils.hashfunc import sha1_hash
# from text_dedup.utils.hashfunc import xxh3_hash
# from text_dedup.utils.timer import Timer
# from text_dedup.utils.tokenization import ngrams
# from text_dedup.utils.union_find import UnionFind

from clean_text_my.text_dedup.utils.add_args import add_bloom_filter_args
from clean_text_my.text_dedup.utils.add_args import add_exact_hash_args
from clean_text_my.text_dedup.utils.add_args import add_io_args
from clean_text_my.text_dedup.utils.add_args import add_meta_args
from clean_text_my.text_dedup.utils.add_args import add_minhash_args
from clean_text_my.text_dedup.utils.add_args import add_sa_args
from clean_text_my.text_dedup.utils.add_args import add_simhash_args
from clean_text_my.text_dedup.utils.hashfunc import sha1_hash
from clean_text_my.text_dedup.utils.hashfunc import xxh3_hash
from clean_text_my.text_dedup.utils.timer import Timer
from clean_text_my.text_dedup.utils.tokenization import ngrams
from clean_text_my.text_dedup.utils.union_find import UnionFind

__all__ = [
    "add_bloom_filter_args",
    "add_exact_hash_args",
    "add_io_args",
    "add_meta_args",
    "add_minhash_args",
    "add_sa_args",
    "add_simhash_args",
    "Timer",
    "ngrams",
    "UnionFind",
    "sha1_hash",
    "xxh3_hash",
]


"""Text deduplication simplified."""

import logging

from rich.logging import RichHandler

logger = logging.getLogger("text_dedup")
logger.setLevel(logging.INFO)
logger.addHandler(RichHandler(rich_tracebacks=True))
logger.propagate = False


import logging
import sys
from logging.handlers import TimedRotatingFileHandler

FORMATTER = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
LOG_FILE = ""


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_file_handler():
    file_handler = TimedRotatingFileHandler(LOG_FILE, when="midnight")
    file_handler.setFormatter(FORMATTER)
    return file_handler


def get_logger(logger_name="clean_text_my"):
    logger = logging.getLogger(logger_name)

    logger.setLevel(logging.DEBUG)

    logger.addHandler(get_console_handler())
    # logger.addHandler(get_file_handler())
    logger.propagate = False

    return logger

logger_name="clean_text_my"
logger = logging.getLogger(logger_name)
logger.setLevel(logging.DEBUG)
logger.addHandler(get_console_handler())
logger.propagate = False

# TODO make user can override the confige



import time


class TimerContext:
    def __init__(self, timer: "Timer", name: str):
        self.timer = timer
        self.name = name
        self.start_time = None

    def __enter__(self):
        self.start_time = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if any([exc_type, exc_val, exc_tb]):
            raise exc_val
        self.timer.elapsed_times[self.name] = time.time() - self.start_time


class Timer:
    """
    A simple timer that tracks the elapsed time of each context.

    Examples
    --------
    >>> t = Timer()
    >>> with t("test"):
    ...     time.sleep(1)
    >>> assert int(t.elapsed_times.get("test", 0)) >= 1, "The elapsed time should be 1 second."
    """

    def __init__(self):
        self.elapsed_times = {}

    def __call__(self, name: str) -> TimerContext:
        """
        Create a context with the given name.

        Parameters
        ----------
        name: str
            The name of the context.

        Returns
        -------
        TimerContext
            The context.

        Examples
        --------
        >>> t = Timer()
        >>> with t("test"):
        ...     time.sleep(1)
        >>> assert int(t.elapsed_times.get("test", 0)) == 1, "The elapsed time should be 1 second."
        >>> with t("test2"):
        ...     time.sleep(2)
        >>> assert int(t.elapsed_times.get("test2", 0)) == 2, "The elapsed time should be 2 seconds."
        """
        return TimerContext(self, name)


import os
import json
import sys
import subprocess
import clean_text_my
from clean_text_my.logging import logger
from clean_text_my.utils import create_dir, slash, is_endwith_slash, platform_name
from datasets import Dataset, load_from_disk


def dir_format(path):
    if not is_endwith_slash(path):
        path = f"{path}{slash()}"

    return path


def jsonl_to_datasets(input_folder: str, output_folder: str) -> None:
    dir_format(input_folder)
    dir_format(output_folder)
    create_dir(output_folder)

    try:
        for root, _, files in os.walk(input_folder):
            for filename in files:
                if filename.endswith(".jsonl"):
                    file_path = os.path.join(root, filename)

                    dataset_name = os.path.splitext(filename)[0]
                    try:
                        with open(file_path) as file:
                            data = [json.loads(line) for line in file]

                        data_dict = {"text": [entry["text"] for entry in data]}

                        dataset = Dataset.from_dict(data_dict)

                        output_folder_path = os.path.join(output_folder, dataset_name)
                        dataset.save_to_disk(output_folder_path)

                        logger.info(f"Dataset saved to {output_folder_path}")
                    except Exception as e:
                        logger.error(
                            f"Error creating dataset for file {file_path}: {e}"
                        )
                        continue
    except Exception as e:
        logger.error(f"Error processing folder {input_folder}: {e}")


def data_deduplication(
    input_folder: str,
    output_folder: str,
    batch_size: int = 10000,
    threshold: float = 0.95,
    min_length: int = 1,
    local: bool = True,
) -> None:
    python_exe = sys.executable

    input_folder = dir_format(input_folder)
    output_folder = dir_format(output_folder)
    create_dir(output_folder)

    try:
        dataset_names = os.listdir(input_folder)
        for dataset_name in dataset_names:
            dataset_path = os.path.join(input_folder, dataset_name)

            if os.path.isdir(dataset_path):
                output_dataset_path = f"{output_folder}{dataset_name}"

                # Construct the command for running minhash.py
                if local:
                    command = f"{python_exe} -m clean_text_my.text_dedup.minhash \
                            --path '{dataset_path}' \
                            --split train \
                            --cache_dir ./cache \
                            --output '{output_dataset_path}' \
                            --column text \
                            --batch_size {str(batch_size)} \
                            --threshold {str(threshold)} \
                            --min_length {str(min_length)} \
                            --local"
                else:
                    command = f"{python_exe} -m clean_text_my.text_dedup.minhash \
                            --path '{dataset_path}' \
                            --split train \
                            --cache_dir ./cache \
                            --output '{output_dataset_path}' \
                            --column text \
                            --batch_size {str(batch_size)} \
                            --threshold {str(threshold)} \
                            --min_length {str(min_length)}"

                # Run the minhash.py script using subprocess
                subprocess.run(command, shell=True)

                logger.info(f"Deduplication completed for {dataset_name}")

    except Exception as e:
        logger.error(f"Error during deduplication: {e}")



import os
import re
import json
import logging
from glob import glob
from clean_text_my.logging import logger
from clean_text_my.utils import create_dir, slash, is_endwith_slash, platform_name

http_errors = [
    "400 Bad Request",
    "401 Unauthorized",
    "402 Payment Required",
    "403 Forbidden",
    "404 Not Found",
    "405 Method Not Allowed",
    "406 Not Acceptable",
    "407 Proxy Authentication Required",
    "408 Request Timeout",
    "409 Conflict",
    "410 Gone",
    "411 Length Required",
    "412 Precondition Failed",
    "413 Payload Too Large",
    "414 URI Too Long",
    "415 Unsupported Media Type",
    "416 Range Not Satisfiable",
    "417 Expectation Failed",
    "418 I'm a teapot",
    "421 Misdirected Request",
    "422 Unprocessable Entity",
    "423 Locked",
    "424 Failed Dependency",
    "425 Too Early",
    "426 Upgrade Required",
    "428 Precondition Required",
    "429 Too Many Requests",
    "431 Request Header Fields Too Large",
    "451 Unavailable For Legal Reasons",
    "500 Internal Server Error",
    "501 Not Implemented",
    "502 Bad Gateway",
    "503 Service Unavailable",
    "504 Gateway Timeout",
    "505 HTTP Version Not Supported",
    "506 Variant Also Negotiates",
    "507 Insufficient Storage",
    "508 Loop Detected",
    "510 Not Extended",
    "511 Network Authentication Required",
]

rejected = [
    "Internal Server Error",
    "__NOEDITSECTION__",
    "enter your username and password",
    "forgotten your password",
    "cookies enabled",
    "enable JavaScript in your browser.",
    "The page cannot be displayed",
    "site or edit the error_page",
]


def replace(input_string: str) -> str:
    input_string = replace_multiple(input_string.replace("…", "."))
    input_string = replace_multiple(input_string, pattern=r"\.{6,}", replace="...")
    return input_string


def replace_multiple(input_string: str, pattern=r"\s{6,}", replace="   ") -> str:
    return re.sub(pattern, replace, input_string)


def reject(input_string: str) -> bool:
    rejected.extend(http_errors)
    if any([r in input_string for r in rejected]):
        return True
    return False


def process(input_folder: str):
    if not is_endwith_slash(input_folder):
        input_folder = f"{input_folder}{slash()}"

    FOLDER_POSTPROCESSING = os.path.join(input_folder, "postprocessing")
    FOLDER_POSTPROCESSING_DONE = os.path.join(input_folder, "postprocessing-done")

    create_dir(FOLDER_POSTPROCESSING)
    create_dir(FOLDER_POSTPROCESSING_DONE)

    files_lst = glob(f"{input_folder}*.jsonl")

    for f in files_lst:
        is_success = True

        directory, filename = os.path.split(f)
        filename_split = os.path.splitext(filename)
        new_done_dir = os.path.join(directory, "postprocessing-done")
        new_dir = os.path.join(directory, "postprocessing")

        new_f_done = os.path.join(new_done_dir, filename)
        new_f = os.path.join(new_dir, filename)

        if os.path.exists(new_f_done):
            logger.info(
                f"Postprocessing for {filename_split[0]} already done, skip ..."
            )
            continue

        with open(new_f, "w") as fopen_l:
            with open(f) as fopen:
                for l in fopen:
                    data = json.loads(l)

                    if "text" not in data:
                        logger.info(
                            f"Skip postprocessing for {filename_split[0]}, error (text) key not found in data"
                        )
                        os.remove(new_f)
                        is_success = False
                        break

                    if reject(data["text"]):
                        continue

                    data = replace(data["text"].strip())

                    if len(data) < 3:
                        continue

                    fopen_l.write(f"{json.dumps(data)}\n")
                    fopen_l.flush()

        if is_success:
            with open(new_f_done, "w") as fopen:
                fopen.write("done")

            logger.info(f"Postprocessing completed for {filename_split[0]}")



from __future__ import annotations

import argparse
import gc
import multiprocessing as mp
import os
import pickle
import random
import csv
import re
from collections import defaultdict
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple
import json
import datasets
import numpy as np
from datasets import load_dataset
from datasets import load_from_disk
from tqdm import tqdm

# from text_dedup import logger
# from text_dedup.utils import UnionFind
# from text_dedup.utils import ngrams
# from text_dedup.utils.add_args import add_io_args
# from text_dedup.utils.add_args import add_meta_args
# from text_dedup.utils.add_args import add_minhash_args
# from text_dedup.utils.analysis import optimal_param
# from text_dedup.utils.hashfunc import sha1_hash
# from text_dedup.utils.hashfunc import xxh3_16hash
# from text_dedup.utils.hashfunc import xxh3_32hash
# from text_dedup.utils.timer import Timer

from clean_text_my.text_dedup import logger
from clean_text_my.text_dedup.utils import UnionFind
from clean_text_my.text_dedup.utils import ngrams
from clean_text_my.text_dedup.utils.add_args import add_io_args
from clean_text_my.text_dedup.utils.add_args import add_meta_args
from clean_text_my.text_dedup.utils.add_args import add_minhash_args
from clean_text_my.text_dedup.utils.analysis import optimal_param
from clean_text_my.text_dedup.utils.hashfunc import sha1_hash
from clean_text_my.text_dedup.utils.hashfunc import xxh3_16hash
from clean_text_my.text_dedup.utils.hashfunc import xxh3_32hash
from clean_text_my.text_dedup.utils.timer import Timer

SEED = 42
RNG = np.random.RandomState(SEED)
NON_ALPHA = re.compile("\W", re.UNICODE)
datasets.logging.set_verbosity_error()


def embed_func(
    content: str,
    idx: int,
    *,
    num_perm: int,
    ngram_size: int,
    min_length: int,
    hashranges: List[Tuple[int, int]],
    permutations: np.ndarray,
    hash_func: Callable,
    dtype: type,
    max_hash: np.uint,
    modulo_prime: np.uint,
) -> Dict[str, Any]:
    """
    Calculate hash values for the content.

    Parameters
    ----------
    content : str
        The content to be embedded.
    idx : int
        The index of the content.
    num_perm : int
        The number of permutations.
    ngram_size : int
        The size of n-grams.
    min_length : int
        The minimum length of the document in terms of tokens.
    hashranges : List[Tuple[int, int]]
        The ranges of hash values.
    permutations : np.ndarray
        The permutations for the minhash.
    hash_func : Callable
        The hash function to use.

    Returns
    -------
    Dict[str, Any]
        The hash values in each range and the index.

    Examples
    --------
    >>> content = "hello world"
    >>> idx = 0
    >>> num_perm = 250
    >>> ngram_size = 1
    >>> hashranges = [(i, i + 25) for i in range(0, 250, 25)]
    >>> max_hash = np.uint32((1 << 32) - 1)
    >>> modulo_prime = np.uint32((1 << 32) - 5)
    >>> PERMUTATIONS = np.array(
    ...     [
    ...         (
    ...             RNG.randint(1, np.uint32((1 << 32) - 5), dtype=np.uint32),
    ...             RNG.randint(0, np.uint32((1 << 32) - 5), dtype=np.uint32),
    ...         )
    ...         for _ in range(num_perm)
    ...     ],
    ...     dtype=np.uint32,
    ... ).T
    >>> res = embed_func(content, idx, num_perm=num_perm, ngram_size=ngram_size, min_length=0, hashranges=hashranges,
    ... permutations=PERMUTATIONS, hash_func=xxh3_32hash,dtype=np.uint32, max_hash=max_hash, modulo_prime=modulo_prime)
    >>> len(res["__signatures__"])
    10
    >>> res["__id__"]
    0
    """
    a, b = permutations
    tokens: Set[bytes] = {
        bytes(" ".join(t).lower(), "utf-8") for t in ngrams(NON_ALPHA.split(content), ngram_size, min_length)
    }

    hashvalues: np.ndarray = np.array([hash_func(token) for token in tokens], dtype=dtype)
    # Permute the hash values to produce new universal hashes
    # Tiling 'a' to match the shape of 'hashvalues'
    # Element-wise multiplication of 'hashvalues' with tiled 'a'
    # Adding 'b' and taking the result modulo 'MERSENNE_PRIME'
    # Performing bitwise AND with 'MAX_HASH'
    hashvalues = np.bitwise_and(
        np.mod(np.add(np.multiply(hashvalues, np.tile(a, (len(hashvalues), 1)).T).T, b), modulo_prime),
        max_hash,
    )
    # this part is where the name "min" of minhash comes from
    # this stacks all the hashes and then takes the minimum from each column
    masks: np.ndarray = np.full(shape=num_perm, dtype=dtype, fill_value=max_hash)
    hashvalues = np.vstack([hashvalues, masks]).min(axis=0)
    # Originally, byteswap was done for speed. Testing show it has a negligible impact
    # keeping  for backward compatibility, even though theoretically and empirically
    # it doesnt matter if it is there or not. github.com/ekzhu/datasketch/issues/114
    Hs = [bytes(hashvalues[start:end].byteswap().data) for start, end in hashranges]
    return {"__signatures__": Hs, "__id__": idx}


if __name__ == "__main__":  # pragma: no cover
    parser = argparse.ArgumentParser(
        prog="text_dedup.minhash",
        description="Deduplicate text using minhash",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = add_io_args(parser)
    parser = add_meta_args(parser)
    parser = add_minhash_args(parser)
    args = parser.parse_args()

    # if not os.path.exists(args.output):
    os.mkdir(args.output)

    HASH_BITS: int = args.hash_bits

    # mypy typing with numpy is difficult
    # 64 bit config is backwards compatibility mode.
    # it uses 64 bit types but almost entirely 32bit data, except for one mersenne prime 2^61
    # why legacy implementations used mersenne primes for modulo:
    # https://en.wikipedia.org/wiki/Universal_hashing#Hashing_strings
    HASH_CONFIG: Dict[int, Tuple[type, Any, Any]] = {
        64: (np.uint64, np.uint64((1 << 32) - 1), np.uint64((1 << 61) - 1)),
        # 32, 16 bit config does not use a mersenne prime.
        # The original reason for using mersenne prime was speed.
        # Testing reveals, there is no benefit to using a 2^61 mersenne prime for division
        32: (np.uint32, np.uint32((1 << 32) - 1), np.uint32((1 << 32) - 5)),
        16: (np.uint16, np.uint16((1 << 16) - 1), np.uint16((1 << 16) - 15)),
    }

    # defaults to backwards compatible HASH_BITS = 64, which is np.uint64 dtypes with 32bit hashes
    DTYPE, MAX_HASH, MODULO_PRIME = HASH_CONFIG.get(HASH_BITS, HASH_CONFIG[64])

    match args.hash_func:
        case "sha1":

            def hash_func(byte_data):
                return sha1_hash(byte_data, d=min(HASH_BITS, 32))

        case "xxh3":
            if HASH_BITS == 16:
                hash_func = xxh3_16hash
            else:
                hash_func = xxh3_32hash

    mp.set_start_method("fork", force=True)
    uf = UnionFind()
    timer = Timer()

    if args.b is not None and args.r is not None:
        B, R = args.b, args.r
    else:
        # Compute the optimal `MinHashLSH` parameter that minimizes the weighted sum
        # of probabilities of false positive and false negative, taken from datasketch.
        # You can also refer to the interactive demo at https://huggingface.co/spaces/bigcode/near-deduplication.
        # The following assumes a "perfect hash". using 16 bit hashes might challenge this assumption
        # lower precision dtype will cause more collisions, so higher false_positives and less false negatives.
        # Both effects move the result towards more documents being considered duplicates.
        B, R = optimal_param(args.threshold, args.num_perm, false_positive_weight=0.5, false_negative_weight=0.5)

    HASH_RANGES = [(i * R, (i + 1) * R) for i in range(B)]
    HASH_TABLES: List[Dict[int, Set]] = [defaultdict(set) for _ in range(B)]

    with timer("Total"):
        with timer("Loading"):
            if args.local:
                ds = load_from_disk(args.path)
            else:
                print('data from huggingface hub')
                ds = load_dataset(
                    path=args.path,
                    name=args.name,
                    data_dir=args.data_dir,
                    data_files=args.data_files,
                    split=args.split,
                    revision=args.revision,
                    cache_dir=args.cache_dir,
                    num_proc=os.cpu_count(),
                    token=args.use_auth_token,
                )

        DATA_SIZE = len(ds)
        PERMUTATIONS: np.ndarray = np.array(
            [
                (
                    RNG.randint(1, MODULO_PRIME, dtype=DTYPE),
                    RNG.randint(0, MODULO_PRIME, dtype=DTYPE),
                )
                for _ in range(args.num_perm)
            ],
            dtype=DTYPE,
        ).T

        with timer("MinHashing"):
            embedded = ds.map(
                function=embed_func,
                fn_kwargs={
                    "num_perm": args.num_perm,
                    "hashranges": HASH_RANGES,
                    "ngram_size": args.ngram,
                    "min_length": args.min_length,
                    "permutations": PERMUTATIONS,
                    "hash_func": hash_func,
                    "dtype": DTYPE,
                    "max_hash": MAX_HASH,
                    "modulo_prime": MODULO_PRIME,
                },
                input_columns=[args.column],
                remove_columns=ds.column_names,
                num_proc=os.cpu_count(),
                with_indices=True,
                desc="Fingerprinting...",
            )

        with timer("Clustering"):
            for i in tqdm(
                range(0, len(embedded), args.batch_size),
                dynamic_ncols=True,
                desc="Iterating MinHashes...",  # noqa: E501
            ):
                batch = embedded[i : i + args.batch_size]
                for key, Hs in zip(batch["__id__"], batch["__signatures__"]):
                    for i, H in enumerate(Hs):
                        HASH_TABLES[i][H].add(key)

            for table in tqdm(HASH_TABLES, dynamic_ncols=True, desc="Clustering..."):
                for cluster in table.values():
                    if len(cluster) <= 1:
                        continue
                    idx = min(cluster)
                    for x in cluster:
                        uf.union(x, idx)

        with timer("Filtering"):
            gc.freeze()
            gc.disable()
            # Dictionary to store duplicate pairs
            duplicate_pairs = {}
            # Open the CSV file for writing
            csvfile = open(f"{args.output}/duplicates_pair.csv", "w", newline="")
            csvwriter = csv.writer(csvfile)
            csvwriter.writerow(["Cluster", "Duplicate Pair"])
   
                            
            ds = ds.map(
                function=lambda _, idx: {"__cluster__": uf.find(idx)},
                with_indices=True,
                num_proc=os.cpu_count(),
                new_fingerprint=str(random.getrandbits(128)),
                desc="Finding clusters...",
            )
        
            # Print duplicate pairs before filtering
            print("Duplicate pairs before filtering:")
            for idx, record in enumerate(ds):
                cluster = record["__cluster__"]
                if cluster in duplicate_pairs:
                    duplicate_pairs[cluster].append(idx)
                else:
                    duplicate_pairs[cluster] = [idx]
            for cluster, indices in duplicate_pairs.items():
                if len(indices) > 1:
                    print(f"Cluster {cluster} has duplicate pairs: {indices}")
                    with open(f"{args.output}/duplicates_cluster.jsonl", "a") as f:
                        for idx in indices:
                            json.dump(ds[idx], f)
                            f.write('\n')
                    duplicate_pair = indices
                    csvwriter.writerow([cluster, duplicate_pair])


            gc.enable()
            gc.collect()
            # This is where the deduplication happens
            # Since there is no easy groupby in datasets
            # I will use this simple filter for now
            final_data = ds.filter(
                function=lambda record, idx: record["__cluster__"] == idx,
                with_indices=True,
                num_proc=os.cpu_count(),
                desc="Filtering clusters...",
            )
  
                    
        with timer("Saving"):
            final_data = final_data.remove_columns(["__cluster__"])
            final_data.save_to_disk(args.output)
            final_data.to_json(f"{args.output}.jsonl")
            if args.debug:
                with open(os.path.join(args.output, "uf.pkl"), "wb") as f:
                    pickle.dump(uf, f, protocol=pickle.HIGHEST_PROTOCOL)

        with timer("Cleaning"):
            if args.clean_cache:
                ds.cleanup_cache_files()
    
              

            
    PAD = 32
    for k, v in timer.elapsed_times.items():
        logger.info(f"{k:<{PAD}}: {v:.2f}s")

    logger.info(f"{'Before':<{PAD}}: {len(ds)}")
    logger.info(f"{'After':<{PAD}}: {len(final_data)}")


from setuptools import setup, find_packages

__packagename__ = "clean_text_my"

with open("requirements.txt") as fopen:
    req = list(filter(None, fopen.read().split("\n")))


def readme():
    # TODO https://github.com/huseinzol05/malaya/blob/master/README-pypi.rst
    with open("README.md") as f:
        return f.read()


setup(
    name=__packagename__,
    description="",
    version="0.1.1",
    long_description=readme(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=req,
    author="Malaysia AI",
    author_email="malaysia.ai2020@gmail.com",
    url="https://github.com/malaysia-ai/clean_text_my",
)



import argparse


def add_io_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:  # pragma: no cover
    """
    Add input/output arguments to parser.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser to add arguments to.

    Returns
    -------
    parser : argparse.ArgumentParser
        Parser with added arguments.
    """
    parser.add_argument("--path", type=str, help="`path` in load_dataset", required=True),
    parser.add_argument("--name", type=str, help="`name` in load_dataset"),
    parser.add_argument("--data_dir", type=str, help="`data_dir` in load_dataset"),
    parser.add_argument("--data_files", type=str, help="`data_files` in load_dataset"),
    parser.add_argument("--split", type=str, help="`split` in load_dataset"),
    parser.add_argument("--cache_dir", type=str, help="`cache_dir` in load_dataset", default=".cache"),
    parser.add_argument("--revision", type=str, help="`revision` in load_dataset"),
    parser.add_argument(
        "--use_auth_token", action=argparse.BooleanOptionalAction, help="To use auth token in load_dataset from HF Hub"
    ),
    parser.add_argument("--local", action=argparse.BooleanOptionalAction, help="Use local dataset", default=False),
    parser.add_argument("--output", type=str, help="Path to deduplicated dataset output", required=True),
    parser.add_argument(
        "--debug", action=argparse.BooleanOptionalAction, help="Whether to run in debug mode", default=False
    )
    parser.add_argument(
        "--clean_cache", action=argparse.BooleanOptionalAction, help="Whether to remove all cache files", default=True
    )
    return parser


def add_meta_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:  # pragma: no cover
    """
    Add meta arguments to parser.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser to add arguments to.

    Returns
    -------
    parser : argparse.ArgumentParser
        Parser with added arguments.
    """
    parser.add_argument(
        "--column",
        type=str,
        help="""Text column to use for deduplication. Concatenate desired columns beforehand if needed.""",
        required=True,
    ),
    parser.add_argument(
        "--batch_size",
        type=int,
        help="""Batch size to use for dataset iteration. Mainly for memory efficiency.
        Single-threaded dedups like exacthash especially benefit from higher batches.
        Batching process itself can take a lot of time. """,
        default=10000,
    ),
    return parser


def add_minhash_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:  # pragma: no cover
    """
    Add MinHash arguments to parser.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser to add arguments to.

    Returns
    -------
    parser : argparse.ArgumentParser
        Parser with added arguments.
    """
    parser.add_argument(
        "--ngram",
        type=int,
        default=5,
        help="Ngram size to use in MinHash.",
    )
    parser.add_argument(
        "--min_length",
        type=int,
        default=5,
        help="Minimum number of tokens to use in MinHash. Shorter documents will be filtered out.",
    )
    parser.add_argument("--seed", type=int, default=42, help="Seed to use in MinHash")
    parser.add_argument("--num_perm", type=int, default=256, help="Number of permutations to use in MinHash")
    parser.add_argument(
        "--threshold", type=float, default=0.7, help="Jaccard similarity threshold to use in MinHashLSH"
    )
    parser.add_argument(
        "--b",
        type=int,
        default=None,
        help="Number of bands",
    )
    parser.add_argument(
        "--r",
        type=int,
        default=None,
        help="Number of rows per band",
    )
    parser.add_argument(
        "--hash_func",
        type=str,
        choices=["sha1", "xxh3"],
        default="sha1",
        help="Hashing algorithm. Defaults to sha1. xxh3 is faster",
    )
    parser.add_argument(
        "--hash_bits",
        type=int,
        choices=[16, 32, 64],
        default=64,
        help="""uint bit precision for hash. default is (np.uint)64.
        However, even when using 64bit precision, only 32 bits are extracted from hash.
        this is due to legacy reasons. refer to ekzhu/datasketch#212.
        """,
    )

    return parser


def add_simhash_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:  # pragma: no cover
    """
    Add SimHash arguments to parser.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser to add arguments to.

    Returns
    -------
    parser : argparse.ArgumentParser
        Parser with added arguments.
    """
    parser.add_argument(
        "--ngram",
        type=int,
        default=3,
        help="""Ngram size to use in SimHash.""",
    )
    parser.add_argument("--f", type=int, default=64, choices=[64, 128], help="Simhash bit size"),
    parser.add_argument("--bit_diff", type=int, default=3, help="Bit difference to use in SimHash"),
    parser.add_argument(
        "--num_bucket", type=int, default=4, help="Number of buckets to use in SimHash, must be larger than bit_diff"
    ),
    return parser


def add_sa_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:  # pragma: no cover
    """
    Add Suffix Array arguments to parser. This adds the following arguments:

    - k: Minimum byte length of a duplicate substring in Suffix Array Deduplication, default 100
    - strategy: Strategy when there are overlapping duplicate substrings, default "overlapping"
        overlapping: Merge all overlapping duplicate substrings
        longest: Only keep the longest duplicate substring
    - google_repo_path: Path to google-research-deduplication codebase, required

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser to add arguments to.

    Returns
    -------
    parser : argparse.ArgumentParser
        Parser with added arguments.
    """
    parser.add_argument(
        "--k", type=int, default=100, help="Minimum byte length of a duplicate substring in Suffix Array Deduplication"
    ),
    parser.add_argument(
        "--strategy",
        type=str,
        default="overlapping",
        help="Strategy when there are overlapping duplicate substrings",
        choices=["overlapping", "longest"],
    )
    parser.add_argument(
        "--google_repo_path", type=str, help="Path to google-research-deduplication codebase", required=True
    ),
    return parser


def add_bloom_filter_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:  # pragma: no cover
    """
    Add Bloom Filter arguments to parser.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser to add arguments to.

    Returns
    -------
    parser : argparse.ArgumentParser
        Parser with added arguments.
    """
    parser.add_argument("--error_rate", type=float, default=1e-6, help="Error rate to use in BloomFilter"),
    parser.add_argument(
        "--hash_func",
        type=str,
        choices=["md5", "sha256", "xxh3"],
        default="md5",
        help="Hash function to use in BloomFilter",
    ),
    parser.add_argument("--initial_capacity", type=int, default=100, help="Initial capacity of BloomFilter"),
    return parser


def add_exact_hash_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:  # pragma: no cover
    """
    Add Exact Hash arguments to parser.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser to add arguments to.

    Returns
    -------
    parser : argparse.ArgumentParser
        Parser with added arguments.
    """
    parser.add_argument(
        "--hash_func",
        type=str,
        choices=["md5", "sha256", "xxh3"],
        default="md5",
        help="Hash function to use in ExactHash. defaults to md5",
    ),
    return parser


import os
import json
import requests
from clean_text_my.logging import logger
from clean_text_my.utils import create_dir, is_endwith_slash, slash


def json_download(url: list, output_folder: str):
    if not is_endwith_slash(output_folder):
        output_folder = f"{output_folder}{slash()}"

    create_dir(output_folder)

    if not isinstance(url, list):
        raise Exception("url must in list")

    for l in url:
        response = requests.get(l)
        save_to = os.path.join(output_folder, l.rsplit("/", 1)[-1])

        open(save_to, "wb").write(response.content)
        logger.info(f"Download completed for {l}")
